-- modified by Einar Mogen

-- modex.e
-- Mode X Graphics Library
-- Pete Eberlein <xseal@harborside.com>
-- July 2, 1998

-- todo:
-- finish page flip (split virtual copy) and then test,test,test

without warning
without type_check

global constant allowed_widths = {256,320,360,376,400}
global constant allowed_heights = {200,224,240,256,270,300,360,400,480,564,600}

-----------------------------------------------------------------
--LOW-LEVEL STUFF REDEFINED--------------------------------------
-----------------------------------------------------------------

global procedure restore_graphics()
    atom junk
    junk = machine_func(5, -1)  -- restore graphics mode
end procedure

global function alloc(integer size)  -- safe allocate
    atom handle
    handle = machine_func(16, size)  -- allocate
    if handle then
        return handle
    end if
    restore_graphics()
    printf(1, "Unable to allocate %d bytes of memory...\nProgram aborted.\n", {size})
    abort(1)
end function

global procedure un_alloc(atom handle) -- free
    machine_proc(17, handle)
end procedure

function allocate_proc(sequence data)
    atom handle
    handle = machine_func(16, length(data))  -- allocate
    if handle then
        poke(handle, data)
        return handle
    end if
    restore_graphics()
    printf(1, "Unable to allocate memory for length-%d sequence...\nProgram aborted.\n", {length(data)})
    abort(1)
end function

function ItoB(atom x)  -- int to 4 bytes
    return and_bits(#FF, floor(x / {1,#100,#10000,#1000000}))
end function

constant in_byte_result = allocate_proc({0})
constant in_byte_proc = allocate_proc(
   {#50,#52,#BA,#00,#00,#00,#00,#EC,#A2}&ItoB(in_byte_result)&{#5A,#58,#C3})
constant out_byte_proc = allocate_proc(
   {#50,#52,#BA,#00,#00,#00,#00,#B0,#00,#EE,#5A,#58,#C3})
constant out_word_proc = allocate_proc(
   {#50,#52,#BA,#00,#00,#00,#00,#B8,#00,#00,#00,#00,#66,#EF,#5A,#58,#C3})

function in_byte(integer port)
    poke4(in_byte_proc + 3, port)
    call(in_byte_proc)
    return peek(in_byte_result)
end function

procedure out_byte(integer port, object data)
    poke4(out_byte_proc + 3, port)
    if atom(data) then
        data = {data}
    end if
    for i = 1 to length(data) do
        poke(out_byte_proc + 8, data[i])
        call(out_byte_proc)
    end for
end procedure

procedure out_word(integer port, object data)
    poke4(out_word_proc + 3, port)
    if atom(data) then
        data = {data}
    end if
    for i = 1 to length(data) do
        poke4(out_word_proc + 8, data[i])
        call(out_word_proc)
    end for
end procedure

-- returns machine codes for "ebx = ebx + edx * width"
function fast_asm_mul(integer width)
    integer count  sequence s
    count = 0      s = {}
    while width do
        if and_bits(width, 1) then
            if count then
                s = s & {#C1,#E2,count} --shl edx, count
            end if
            s = s & {#01,#D3}       --add ebx, edx
            count = 0
        end if
        count = count + 1
        width = floor(width / 2)
    end while
    return s
end function

-----------------------------------------------------------------
--SCREENS--------------------------------------------------------
-----------------------------------------------------------------

global constant screen = alloc(24)
mem_set(screen, 0, 24)

atom current_source, current_dest
current_source = screen
current_dest = screen
integer dest_width, dest_height -- used for clipping in pixel()
    
constant video_mem = #A0000
constant sequ = #3C4,  -- sequencer register base
         crtc = #3D4,  -- CRT controller register base
         misc = #3C2,  -- miscellaneous output register
         ins1 = #3DA,  -- input status #1 register
         attr = #3C0   -- attribute register 


constant --offsets for data in a screen handle
    screen_width = 0,
    screen_height = 4,
    screen_source = 8,
    screen_dest = 12,
    screen_clip_width = 16,
    screen_clip_height = 20,
    screen_planes = 24   

-- creates a virtual screen and returns its handle
global function virtual_screen(integer width, integer height)
    atom screen_mem, handle
    sequence source_code

    width = floor((width+3)/4)
    source_code = 
        fast_asm_mul(width*4)&
       {#88,#DA,                --mov dl, bl
        #83,#E2,#03,            --and edx, 3
        #C1,#E2,#02,            --shl edx, 2
        #C1,#FB,#02,            --sar ebx, 2
        #03,#9A,#00,#00,#00,#00,--add ebx, dword ptr [video_mem + edx]
        #C3}                    --ret
    handle = alloc(40 + length(source_code) + 4*width*height)
    screen_mem = handle + 40 + length(source_code)
    poke4(handle + screen_width, width*4)
    poke4(handle + screen_height, height)
    poke4(handle + screen_source, handle + 40)
    poke4(handle + screen_dest, handle + 40)
    poke4(handle + screen_clip_width, width*4)
    poke4(handle + screen_clip_height, height)
    poke4(handle + screen_planes, screen_mem + height*width*{0,1,2,3})
    poke(handle + 40, source_code)
    poke4(screen_mem - 5,  handle + screen_planes)
    mem_set(screen_mem, 0, 4*width*height)

    return handle
end function

global function sub_screen(atom parent_screen, integer width, integer height, object offset)
    atom handle
    handle = allocate_proc(
   {#00,#00,#00,#00,        --    0: dd  ?
    #00,#00,#00,#00,        --    4: dd  ?
    #00,#00,#00,#00,        --    8: dd  ?
    #00,#00,#00,#00,        --    C: dd  ?
    #00,#00,#00,#00,        --   10: dd  ?
    #00,#00,#00,#00,        --   14: dd  ?
    #81,#C3,#00,#00,#00,#00,--   18: add ebx, source_add (26)
    #FF,#25,#00,#00,#00,#00,--   1E: jmp near dword ptr [source_proc] (32)
    #81,#C3,#00,#00,#00,#00,--   24: add ebx, dest_add (38)
    #FF,#25,#00,#00,#00,#00}--   2A: jmp near dword ptr [dest_proc] (44)
    )
    poke4(handle, peek4u({parent_screen + screen_width, 2}) & 
        {handle + #18, handle + #24, width, height})

    if sequence(offset) then
        offset = offset[1] + offset[2] * peek4u(parent_screen + screen_width)
    end if
    poke4(handle + 26, offset)
    poke4(handle + 32, parent_screen + screen_source)
    poke4(handle + 38, offset)
    poke4(handle + 44, parent_screen + screen_dest)
    
    return handle
end function


-----------------------------------------------------------------
--MODES----------------------------------------------------------
-----------------------------------------------------------------

-- sets the virtual width of the screen
global procedure set_width(integer width)
    out_byte(crtc, #13)
    width = floor((width + 7) / 8)
    out_byte(crtc+1, width)
    poke(#44A, remainder(width, 256) & floor(width / 256)) -- BIOS Video Columns
    width = width * 2

    poke4(screen + screen_width, width*4)
    poke4(screen + screen_clip_width, width*4)

    -- initialize the screen handle
    if peek4u(screen + screen_source) then
        un_alloc(peek4u(screen + screen_source))
    end if
    if peek4u(screen + screen_dest) then
        un_alloc(peek4u(screen + screen_dest))
    end if

    poke4(screen + screen_source, allocate_proc(
              {#50,                    --push eax
--               #81,#C3,#00,#00,#00,#00,--add ebx, dword source_add
               #88,#DC,                --mov ah, bl
               #80,#E4,#03,            --and ah, 3
               #C1,#FB,#02,            --sar ebx, 2
               #81,#C3}&ItoB(video_mem)&--add ebx, dword video_mem (34)
               fast_asm_mul(width)&{   --ebx = ebx + edx * width
               #BA,#CE,#03,#00,#00,    --mov edx, #3CE
               #B0,#04,                --mov al, 4
               #66,#EF,                --out dx, ax
               #58,                    --pop eax
               #C3}                    --ret
            )
        )
    poke4(screen + screen_dest, allocate_proc(
              {#50,                    --push eax
               #51,                    --push ecx
--               #81,#C3,#00,#00,#00,#00,--add ebx, dword dest_add
               #88,#D9,                --mov cl, bl
               #80,#E1,#03,            --and cl, 3
               #C1,#FB,#02,            --sar ebx, 2
               #81,#C3}&ItoB(video_mem)&--add ebx, dword video_mem
               fast_asm_mul(width)&{   --ebx = ebx + edx * width
               #BA,#C4,#03,#00,#00,    --mov edx, #3C4
               #B4,#01,                --mov ah, 1
               #B0,#00,                --mov al, 0
               #D3,#E0,                --shl eax, cl
               #B0,#02,                --mov al, 2
               #66,#EF,                --out dx, ax
               #59,                    --pop ecx
               #58,                    --pop eax
               #C3}                    --ret
            )
        )

end procedure

global function mode_x(integer width, integer height)
    sequence h_regs, v_regs
    integer misc_value
    misc_value = #23
    if width = 256 then
        h_regs = {#5F00,#3F01,#4002,#8203,#4A04,#9A05,#2013}
        misc_value = misc_value + #00
    elsif width = 320 then
        h_regs = {#5F00,#4F01,#5002,#8203,#5404,#8005,#2813}
        misc_value = misc_value + #00
    elsif width = 360 then
        h_regs = {#6B00,#5901,#5A02,#8E03,#5E04,#8A05,#2D13}
        misc_value = misc_value + #04
    elsif width = 376 then
        h_regs = {#6E00,#5D01,#5E02,#9103,#6204,#8F05,#2F13}
        misc_value = misc_value + #04
    elsif width = 400 then
        h_regs = {#7100,#6301,#6402,#9203,#6504,#8205,#3213}
        misc_value = misc_value + #04
    else
        return 1  -- unsupported width
    end if
    if height = 200 then
        v_regs = {#BF06,#1F07,#4109,#9C10,#8E11,#8F12,#9615,#B916}
        misc_value = misc_value + #40
    elsif height = 224 then
        v_regs = {#0B06,#3E07,#4109,#DA10,#9C11,#BF12,#C715,#0416}
        misc_value = misc_value + #C0
    elsif height = 240 then
        v_regs = {#0D06,#3E07,#4109,#EA10,#AC11,#DF12,#E715,#0616}
        misc_value = misc_value + #C0
    elsif height = 256 then
        v_regs = {#2306,#B207,#6109,#0A10,#AC11,#FF12,#0715,#1A16}
        misc_value = misc_value + #C0
    elsif height = 270 then
        v_regs = {#3006,#F007,#6109,#2010,#A911,#1B12,#1F15,#2F16}
        misc_value = misc_value + #C0
--    elsif height = 282 then
--        v_regs = {#6206,#0F07,#6109,#3710,#8911,#3312,#3C15,#5C16}
--        misc_value = misc_value + #C0
    elsif height = 300 then
        v_regs = {#4606,#1F07,#4009,#3110,#8011,#2B12,#2F15,#4416}
        misc_value = misc_value + #00
--    elsif height = 308 then
--        v_regs = {#6206,#0F07,#4009,#3710,#8911,#3312,#3C15,#5C16}
--        misc_value = misc_value + #C0
    elsif height = 360 then
        v_regs = {#BF06,#1F07,#4009,#8810,#8511,#6712,#6D15,#BA16}
        misc_value = misc_value + #40
    elsif height = 400 then
        v_regs = {#BF06,#1F07,#4009,#9C10,#8E11,#8F12,#9615,#B916}
        misc_value = misc_value + #40
    elsif height = 480 then
        v_regs = {#0D06,#3E07,#4009,#EA10,#AC11,#DF12,#E715,#0616}
        misc_value = misc_value + #C0
    elsif height = 564 then
        v_regs = {#6206,#F007,#6009,#3710,#8911,#3312,#3C15,#5C16}
        misc_value = misc_value + #C0
    elsif height = 600 then
        v_regs = {#7006,#F007,#6009,#5B10,#8C11,#5712,#5815,#7016}
        misc_value = misc_value + #C0
    else 
        return 1  -- unsupported height
    end if

    if machine_func(5, 19) then  -- attempt graphics mode 19
        return 1
    end if
        
    -- set planar mode
    out_word(sequ, #0604)

    -- clear the screen
    out_word(sequ, #0F02)
    mem_set(video_mem, 0, #10000)

    -- more to set the planar mode
    out_word(sequ, #0100)
    out_byte(misc, misc_value)
    out_word(sequ, #0300)

    -- allow access to CRTC protected registers
    out_byte(crtc, #11)
    out_byte(crtc+1, and_bits(in_byte(crtc+1), #7F))

    -- set horizontal and vertical registers
    out_word(crtc, h_regs & v_regs & {#0008,#0014,#E317})

    -- set the overscan color to 0
    if in_byte(ins1) then end if
    out_byte(attr, {#31, 0})

    set_width(width)
    poke4(screen + screen_height, floor(65536 / width))
    poke4(screen + screen_clip_height, height)

    return 0  -- success
end function


constant 
    get_addr_result = alloc(4),
    get_addr_proc = allocate_proc(
   {#60,                    --    0: pusha
    #BB,#00,#00,#00,#00,    --    1: mov ebx, x (2)
    #BA,#00,#00,#00,#00,    --    6: mov edx, y (7)
    #FF,#15,#00,#00,#00,#00,--    B: call near dword ptr [handleofs] (13)
    #89,#1D,#00,#00,#00,#00,--   11: mov [result], ebx (19)
    #61,                    --   17: popa
    #C3})                   --   18: ret
poke4(get_addr_proc + 19, get_addr_result)

function get_source_addr(sequence pos)
    poke4(get_addr_proc + 2, pos[1])
    poke4(get_addr_proc + 7, pos[2])
    poke4(get_addr_proc + 13, current_source + screen_source)
    call(get_addr_proc)
    return peek4u(get_addr_result)
end function

function get_dest_addr(sequence pos)
    poke4(get_addr_proc + 2, pos[1])
    poke4(get_addr_proc + 7, pos[2])
    poke4(get_addr_proc + 13, current_dest + screen_dest)
    call(get_addr_proc)
    return peek4u(get_addr_result)
end function

-----------------------------------------------------------------
--SAVING AND LOADING SCREENS IN A FILE---------------------------
-----------------------------------------------------------------

-- saves the selected source screen to either:
--  - a file named by sequence fn
--  - an already opened binary file number fn
global procedure save_virtual_screen(object fn)
    integer width, height, newfile
    newfile = sequence(fn)
    if newfile then
        fn = open(fn, "wb")
    end if
    width = peek4u(current_source + screen_width)
    height = peek4u(current_source + screen_height)
    puts(fn, ItoB(width) & ItoB(height))
    width = floor(width / 4)
    for i = 0 to 3 do
        puts(fn, peek({get_source_addr({i,0}), width*height}))
    end for
    if newfile then
        close(fn)
    end if
end procedure

-- creates a virtual screen exactly like the one that was saved
global function load_virtual_screen(object fn)
    integer newfile
    object width, height
    sequence data
    newfile = sequence(fn)
    if newfile then
        fn = open(fn, "rb")
    end if
    width = getc(fn)
    width = width + #100 * getc(fn)
    width = width + #10000 * getc(fn)
    width = width + #1000000 * getc(fn)
    height = getc(fn)
    height = height + #100 * getc(fn)
    height = height + #10000 * getc(fn)
    height = height + #1000000 * getc(fn)
    current_dest = virtual_screen(width, height)
    width = floor(width / 4)
    data = repeat(0, width * height)
    for i = 0 to 3 do
        for j = 1 to length(data) do
            data[j] = getc(fn)
        end for
        poke(get_dest_addr({i,0}), data)
    end for
    if newfile then
        close(fn)
    end if
    return current_dest
end function

-----------------------------------------------------------------
-- COMMANDS -----------------------------------------------------
-----------------------------------------------------------------

global atom 
    copy, 
    clear_copy, 
    mix_copy, 
    wait_retrace

constant wait_retrace_code =
    {#66,#BA,#DA,#03,#EC,#A8,#08,#75,#00,#EC,#A8,#08,#74,#FB}

wait_retrace = allocate_proc({#60} & wait_retrace_code & {#61,#C3})


-- offsets for the command structure
constant 
    start_offset = #13,
    dest_x = 3,
    dest_y = 7,
    dest_z = 11, 
    prev_command = 15,
    next_command = 21,
    loop_offset = #14

constant 
    loop_proc = alloc(4),
    return_proc = allocate_proc(
   {#C3,                    --    0: ret
    #90,#90,                --    1: nop nop
    #00,#00,#00,#00,        --    3: dest_x: dd ? (3)
    #00,#00,#00,#00,        --    7: dest_y: dd ? (7)
    #FF,#FF,#FF,#7F,        --    B: dest_z: dd #7FFFFFFF (11)
    #00,#00,#00,#00,        --    F: prev_command: dd ? (15)
    #60,                    --   13: start_offset: pusha
    #B8,#00,#00,#00,#00,    --   14: loop_offset: mov eax, next_command (21)
    #A3}&ItoB(loop_proc)&{  --   19: mov [loop_proc], eax (31)
    #61,                    --   14: popa
    #C3})                   --   15: ret

poke4(loop_proc, return_proc + loop_offset)
poke4(return_proc + next_command, return_proc + loop_offset)

constant command_header =
   {#60,                    --    0: pusha
    #EB,#20,                --    1: jmp short label1
    #00,#00,#00,#00,        --    3: dest_x: dd ? (3)
    #00,#00,#00,#00,        --    7: dest_y: dd ? (7)
    #00,#00,#00,#00,        --    B: dest_z: dd ? (11)
    #00,#00,#00,#00,        --    F: prev_command: dd ? (15)
    #60,                    --   13: start_offset: pusha
    #B8,#00,#00,#00,#00,    --   14: loop_offset: mov eax, next_command (21)
    #05,loop_offset,#00,#00,#00,--   19: add eax, loop_offset (26)
    #A3}&ItoB(loop_proc)    --   1E: mov [loop_proc], eax (31)
                            --   23: label1:

constant generic_copy_code =
    command_header & 
   {#BB,#00,#00,#00,#00,    -- mov ebx, command_handle (1)
    #B8,#00,#00,#00,#00,    -- mov eax, copy_proc (6)
    #FF,#E0,                -- jmp near eax
    0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0}
    
constant command_footer =
   {#FF,#25}&ItoB(loop_proc)--jmp near dword ptr [loop_proc]

constant 
    generic_command_handle = length(command_header) + 1,
    generic_copy_proc      = length(command_header) + 6,
    generic_data = length(command_header) + 12

constant 
    dest_screen = generic_data + 0,
    source_handle = generic_data + 4,
    mix_table = generic_data + 8,
    collision_flags = generic_data + 12,
    source_data = generic_data + 16,

    source_screen = 0,
    source_x = 4,
    source_y = 8,
    source_w = 12,
    source_h = 16,
    source_xoff = 20,
    source_yoff = 24,
    source_next = 28,
    source_size = 32
    
constant scratch_command = allocate_proc(generic_copy_code & repeat(0,source_size))
    poke4(scratch_command + generic_command_handle, scratch_command)


atom first_command
first_command = return_proc


global procedure set_dest_pos(atom handle, sequence pos)
    atom next, prev
    if length(pos) = 3 then
        next = peek4u(handle + next_command)
        prev = peek4u(handle + prev_command)
        if next then
            poke4(next + prev_command, prev)
        end if
        if prev then
            poke4(prev + next_command, next)
        end if
        if handle = first_command then
            first_command = next
        end if
    
        next = first_command
        while pos[3] > peek4s(next + dest_z) do
            next = peek4u(next + next_command)
        end while
        
        if next = first_command then
            poke4(handle + next_command, first_command)
            poke4(handle + prev_command, 0)
            poke4(first_command + prev_command, handle)
            first_command = handle
        else
            prev = peek4u(next + prev_command)
            poke4(prev + next_command, handle)
            poke4(next + prev_command, handle)
            poke4(handle + next_command, next)
            poke4(handle + prev_command, prev)
        end if 
    end if
    poke4(handle + dest_x, pos)
end procedure

sequence copy_procs

global function new_command(atom kind, object source, sequence dest)
    atom handle

    if kind = copy or kind = clear_copy or kind = mix_copy then


        if sequence(source) then
            handle = allocate_proc(generic_copy_code & repeat(0,source_size))
            poke4(handle + source_data, source)
            source = handle + source_data
        else
            handle = allocate_proc(generic_copy_code)
        end if
        poke4(handle + source_handle, source)
        poke4(handle + generic_command_handle, handle)
        poke4(handle + generic_copy_proc, kind)
        poke4(handle + dest_screen, dest[1])
        
    elsif kind = wait_retrace then
        handle = allocate_proc(
            command_header&
            wait_retrace_code& 
            command_footer)
    else
        puts(1, "Error creating command.\n")
        abort(1)
    end if
    set_dest_pos(handle, dest[2..length(dest)])
    return handle
end function

global procedure do_command(atom kind, object source, sequence dest)
    if kind = copy or kind = clear_copy then
        if sequence(source) then
            poke4(scratch_command + source_data, source)
            source = scratch_command + source_data
        end if
        poke4(scratch_command + source_handle, source)
        poke4(scratch_command + generic_copy_proc, kind)
        poke4(scratch_command + dest_screen, dest[1])
        poke4(scratch_command + dest_x, dest[2..3])
        call(scratch_command)
    end if    
end procedure


global procedure clear_commands()
    atom temp
    while first_command != return_proc do
        temp = peek4u(first_command + next_command)
        un_alloc(first_command)
        first_command = temp
    end while        
end procedure

global procedure execute_commands()
    call(first_command + start_offset)
end procedure

global procedure free_command(atom handle)
    atom next, prev
    next = peek4u(handle + next_command)
    prev = peek4u(handle + prev_command)
    if next then
        poke4(next + prev_command, prev)
    end if
    if prev then
        poke4(prev + next_command, next)
    end if
    if handle = first_command then
        first_command = next
    end if
    un_alloc(handle)
end procedure

global function chain_commands(sequence cmds)
    for i = 1 to length(cmds)-1 do
        poke4(cmds[i] + next_command, cmds[i+1])
    end for
    poke4(cmds[length(cmds)] + next_command, return_proc)
    return cmds[1]
end function

global procedure set_source_handle(atom handle, atom a)
    poke4(handle + source_handle, a)
end procedure

global procedure set_source_pos(atom handle, sequence pos)
    poke4(handle + source_x, pos)
end procedure

global procedure set_source_size(atom handle, sequence size)
    poke4(handle + source_w, size)
end procedure

global procedure set_mixtable(atom handle, atom mixtable)
    poke4(handle + mix_table, mixtable)
end procedure

global procedure set_dest(atom handle, atom dest)
    poke4(handle + dest_screen, dest)
end procedure

global procedure select_source(atom source)
    current_source = source
end procedure

global procedure select_dest(atom dest)
    current_dest = dest
    dest_width = peek4u(dest + screen_width)
    dest_height = peek4u(dest + screen_height)
end procedure

global procedure select_both(atom both)
    current_source = both
    current_dest = both
    dest_width = peek4u(both + screen_width)
    dest_height = peek4u(both + screen_height)
end procedure

-- pan/originate proc
constant pan_proc = allocate_proc(
   {#60,                    --    0: pusha
    #BB,#00,#00,#00,#00,    --    1: mov ebx, pan_x (2)
    #BA,#DA,#03,#00,#00,    --mov edx, #3DA
    #EC,                    --in al, dx
    #B2,#C0,                --mov dl, #C0
    #B0,#33,                --mov al, #33
    #EE,                    --out dx, al
    #88,#D8,                --mov al, bl
    #24,#03,                --and al, 3
    #D0,#E0,                --shl al, 1
    #EE,                    --out dx, al
    #C1,#EB,#02,            --shr ebx, 2
    #B2,#D4,                --mov dl, #D4
    #B0,#0C,                --mov al, #0C
    #88,#FC,                --mov ah, bh
    #66,#EF,                --out dx, ax
    #B0,#0D,                --mov al, #0D
    #88,#DC,                --mov ah, bl
    #66,#EF,                --out dx, ax
    #61,                    --   12: popa
    #C3}                    --   13: ret
    )

global procedure pan(sequence pos)
    poke4(pan_proc + 2, pos[1] + pos[2] * peek4u(screen))
    call(pan_proc)
end procedure

global function allocate_source(sequence source_data)
    atom hsource
    hsource = alloc(source_size)
    mem_set(hsource, 0, source_size)
    poke4(hsource, source_data)
    return hsource
end function

global function animated_source(integer loop, sequence frames)
    atom hsource
    hsource = alloc(source_size * length(frames))
    mem_set(hsource, 0, source_size * length(frames))
    for i = 1 to length(frames) do
        poke4(hsource + (i-1)*source_size, frames[i])
        poke4(hsource + (i-1)*source_size + source_next, hsource + i*source_size)
    end for
    if loop > length(frames) then
        loop = 0
    end if
    poke4(hsource + (length(frames)-1)*source_size + source_next, 
        (hsource + loop * source_size) * (loop != 0))
    return hsource
end function

-----------------------------------------------------------------
-- PAGE-FLIPPING ------------------------------------------------
-----------------------------------------------------------------
global constant 
    flip_virtual = 1,
    flip_page = 2
atom frame_rate, time_frame, flip_cmd
global atom flip_vs
flip_vs = screen
integer flip_change, flip_mode, flip_offset

global procedure flip_setup(integer mode, atom scrn, atom rate)
    integer w, h
    if flip_vs != screen then 
        un_alloc(flip_vs) 
        flip_vs = screen 
    end if
    if rate then
        frame_rate = 1 / rate
        time_frame = time() + frame_rate
    else
        frame_rate = 0
    end if
    if scrn then
        w = peek4u(scrn + screen_clip_width)
        h = peek4u(scrn + screen_clip_height)
    end if
    flip_mode = mode
    if mode = flip_virtual then
        flip_vs = virtual_screen(w, h)
        flip_cmd = new_command(copy, {flip_vs,0,0,w,h}, {scrn,0,0})
    elsif mode = flip_page and scrn = screen then
        flip_change =  h * peek4u(scrn + screen_width)
        flip_offset = flip_change
        flip_vs = sub_screen(scrn, w, h, flip_offset)
    else
        pan({0,0})
    end if
end procedure

global procedure flip()
  while time() < time_frame do
  end while

  time_frame = time_frame + frame_rate
        
  if time() > time_frame + 1 then -- reset if over a second behind
    time_frame = time()
  elsif time() > time_frame then  -- return if frame has expired
    return
  end if

    -- execute all commands
  call(first_command + start_offset)

--        pan
  poke4(pan_proc + 2, flip_offset)
  call(pan_proc)
  flip_offset = flip_change - flip_offset
  poke4(flip_vs + 26, flip_offset)
  poke4(flip_vs + 38, flip_offset)
end procedure

-----------------------------------------------------------------
-- PIXEL, GETPIXEL ----------------------------------------------
-----------------------------------------------------------------

global procedure pixel(sequence p, sequence xy)
-- p MUST be a sequence after Einar's "speed-tweak" for hayes
    atom addr
    for j = 1 to 4 do
        poke4(get_addr_proc + 2, xy[1])
        poke4(get_addr_proc + 7, xy[2])
        poke4(get_addr_proc + 13, current_dest + screen_dest)
        call(get_addr_proc)
        addr = peek4u(get_addr_result)
        for i = j to length(p) by 4 do
            poke(addr, p[i])
            addr += 1
        end for
        xy[1] += 1
    end for
end procedure

global function get_pixel(sequence xy)
    atom addr
    sequence result
    if length(xy) = 2 then
        return peek(get_source_addr(xy))
    else
        result = repeat(0, xy[3])
        for j = 1 to 4 do
            addr = get_source_addr(xy)
            for i = j to xy[3] by 4 do
                result[i] = peek(addr)
                addr = addr + 1
            end for
            xy[1] = xy[1] + 1
        end for
        return result
    end if
end function


global procedure line(sequence p1, sequence p2, integer color)
    integer x, y, w, h, dx, dy, count
    atom addr

    x = p1[1]
    y = p1[2]
    dx = p2[1] - x
    dy = p2[2] - y
    if dx < 0 then
        w = -dx+1
    else
        w = dx+1
    end if
    dx = (dx>0)-(dx<0)
    if dy < 0 then
        h = -dy+1
    else
        h = dy+1
    end if
    dy = (dy>0)-(dy<0)
    
    if w + h = 2 then
        pixel(color, p1)
    elsif w >= h then
        count = floor(h/2)
        poke4(get_addr_proc + 13, current_dest + screen_dest)
        poke4(get_addr_proc + 7, y)
        for i = x to p2[1] by dx do
            poke4(get_addr_proc + 2, i)
            call(get_addr_proc)
            poke(peek4u(get_addr_result), color)
            count = count + h
            if count >= w then
                y = y + dy
                poke4(get_addr_proc + 7, y)
                count = count - w
            end if
        end for
    else
        count = floor(w/2)
        poke4(get_addr_proc + 13, current_dest + screen_dest)
        poke4(get_addr_proc + 2, x)
        for i = y to p2[2] by dy do
            poke4(get_addr_proc + 7, i)
            call(get_addr_proc)
            poke(peek4u(get_addr_result), color)
            count = count + w
            if count >= h then
                x = x + dx
                poke4(get_addr_proc + 2, x)
                count = count - h
            end if
        end for
    end if
            
            
            
end procedure


-----------------------------------------------------------------
--INITIALIZATION-------------------------------------------------
-----------------------------------------------------------------
select_source(screen)
select_dest(screen)

copy = allocate_proc({
#8B,#7B,#2F,#89,#3D,#C2,#09,#4D,#00,#8B,#73,#33,#8B,#46,#1C,#85,#C0,#74,#03,#89,
#43,#33,#8B,#06,#A3,#A2,#09,#4D,#00,#8B,#56,#0C,#8B,#43,#03,#8B,#4E,#04,#03,#46,
#14,#7D,#08,#01,#C2,#7E,#64,#29,#C1,#31,#C0,#A3,#C7,#09,#4D,#00,#89,#0D,#A7,#09,
#4D,#00,#F7,#D8,#03,#47,#10,#39,#C2,#7F,#02,#89,#D0,#83,#F8,#00,#7E,#45,#83,#C0,
#03,#A3,#92,#09,#4D,#00,#8B,#56,#10,#8B,#43,#07,#8B,#4E,#08,#03,#46,#18,#7D,#08,
#01,#C2,#7E,#2B,#29,#C1,#31,#C0,#A3,#CE,#09,#4D,#00,#89,#0D,#AE,#09,#4D,#00,#F7,
#D8,#03,#47,#14,#39,#C2,#7F,#02,#89,#D0,#83,#F8,#00,#7E,#0C,#A3,#9D,#09,#4D,#00,
#BD,#04,#00,#00,#00,#EB,#06,#FF,#25,#00,#00,#00,#00,#B9,#00,#00,#00,#00,#C1,#E9,
#02,#74,#F0,#55,#BD,#00,#00,#00,#00,#BE,#00,#00,#00,#00,#BB,#00,#00,#00,#00,#8B,
#06,#BA,#00,#00,#00,#00,#C1,#E8,#02,#FF,#56,#08,#29,#C8,#89,#DE,#A3,#24,#0A,#4D,
#00,#BF,#00,#00,#00,#00,#BB,#00,#00,#00,#00,#8B,#07,#BA,#00,#00,#00,#00,#C1,#E8,
#02,#FF,#57,#0C,#29,#C8,#89,#DF,#A3,#2A,#0A,#4D,#00,#F7,#DB,#81,#E3,#03,#00,#00,
#00,#29,#D9,#7D,#04,#01,#CB,#31,#C9,#89,#1D,#0E,#0A,#4D,#00,#88,#CB,#C1,#E9,#02,
#80,#E3,#03,#89,#0D,#15,#0A,#4D,#00,#89,#1D,#1C,#0A,#4D,#00,#EB,#00,#B9,#00,#00,
#00,#00,#F3,#A4,#B9,#00,#00,#00,#00,#F3,#A5,#B9,#00,#00,#00,#00,#F3,#A4,#81,#C6,
#00,#00,#00,#00,#81,#C7,#00,#00,#00,#00,#4D,#75,#DC,#5D,#FF,#05,#A7,#09,#4D,#00,
#FF,#05,#C7,#09,#4D,#00,#FF,#0D,#92,#09,#4D,#00,#4D,#0F,#85,#46,#FF,#FF,#FF,#FF,
#25,#00,#00,#00,#00})
poke4(copy + 5, copy + 202) -- @dest_vs
poke4(copy + 25, copy + 170) -- @source_vs
poke4(copy + 52, copy + 207) -- @dest_x
poke4(copy + 58, copy + 175) -- @source_x
poke4(copy + 82, copy + 154) -- @source_w
poke4(copy + 109, copy + 214) -- @dest_y
poke4(copy + 115, copy + 182) -- @source_y
poke4(copy + 136, copy + 165) -- @source_h
poke4(copy + 149, loop_proc)
poke4(copy + 197, copy + 300) -- @source_add
poke4(copy + 229, copy + 306) -- @dest_add
poke4(copy + 251, copy + 278) -- @line1
poke4(copy + 265, copy + 285) -- @line2
poke4(copy + 271, copy + 292) -- @line3
poke4(copy + 316, copy + 175) -- @source_x
poke4(copy + 322, copy + 207) -- @dest_x
poke4(copy + 328, copy + 154) -- @source_w
poke4(copy + 341, loop_proc)

clear_copy = allocate_proc({
#8B,#7B,#2F,#89,#3D,#32,#33,#4D,#00,#8B,#73,#33,#8B,#46,#1C,#85,#C0,#74,#03,#89,
#43,#33,#8B,#06,#A3,#12,#33,#4D,#00,#8B,#56,#0C,#8B,#43,#03,#8B,#4E,#04,#03,#46,
#14,#7D,#08,#01,#C2,#7E,#64,#29,#C1,#31,#C0,#A3,#37,#33,#4D,#00,#89,#0D,#17,#33,
#4D,#00,#F7,#D8,#03,#47,#10,#39,#C2,#7F,#02,#89,#D0,#83,#F8,#00,#7E,#45,#83,#C0,
#03,#A3,#02,#33,#4D,#00,#8B,#56,#10,#8B,#43,#07,#8B,#4E,#08,#03,#46,#18,#7D,#08,
#01,#C2,#7E,#2B,#29,#C1,#31,#C0,#A3,#3E,#33,#4D,#00,#89,#0D,#1E,#33,#4D,#00,#F7,
#D8,#03,#47,#14,#39,#C2,#7F,#02,#89,#D0,#83,#F8,#00,#7E,#0C,#A3,#0D,#33,#4D,#00,
#BD,#04,#00,#00,#00,#EB,#06,#FF,#25,#00,#00,#00,#00,#B9,#00,#00,#00,#00,#C1,#E9,
#02,#74,#F0,#55,#BD,#00,#00,#00,#00,#BE,#00,#00,#00,#00,#BB,#00,#00,#00,#00,#8B,
#06,#BA,#00,#00,#00,#00,#C1,#E8,#02,#FF,#56,#08,#29,#C8,#89,#DE,#A3,#A7,#33,#4D,
#00,#BF,#00,#00,#00,#00,#BB,#00,#00,#00,#00,#8B,#07,#BA,#00,#00,#00,#00,#C1,#E8,
#02,#FF,#57,#0C,#29,#C8,#89,#DF,#A3,#AD,#33,#4D,#00,#F7,#DB,#81,#E3,#03,#00,#00,
#00,#29,#D9,#7D,#04,#01,#CB,#31,#C9,#89,#1D,#88,#33,#4D,#00,#88,#CB,#80,#E3,#03,
#C1,#E9,#02,#75,#08,#01,#1D,#88,#33,#4D,#00,#89,#CB,#89,#0D,#9F,#33,#4D,#00,#89,
#1D,#16,#34,#4D,#00,#EB,#00,#B9,#00,#00,#00,#00,#85,#C9,#74,#0D,#8A,#06,#84,#C0,
#74,#02,#88,#07,#46,#47,#49,#75,#F3,#81,#C1,#00,#00,#00,#00,#75,#2F,#81,#C6,#00,
#00,#00,#00,#81,#C7,#00,#00,#00,#00,#4D,#75,#D3,#5D,#FF,#05,#17,#33,#4D,#00,#FF,
#05,#37,#33,#4D,#00,#FF,#0D,#02,#33,#4D,#00,#4D,#0F,#85,#33,#FF,#FF,#FF,#FF,#25,
#00,#00,#00,#00,#8B,#06,#85,#C0,#74,#31,#89,#C3,#C1,#EB,#10,#84,#C0,#74,#10,#84,
#E4,#74,#42,#84,#DB,#74,#43,#84,#FF,#74,#47,#89,#07,#EB,#18,#47,#84,#E4,#74,#02,
#88,#27,#47,#84,#DB,#74,#02,#88,#1F,#47,#84,#FF,#74,#02,#88,#3F,#83,#EF,#03,#83,
#C7,#04,#83,#C6,#04,#49,#75,#C0,#81,#C1,#00,#00,#00,#00,#74,#89,#8A,#06,#84,#C0,
#74,#02,#88,#07,#46,#47,#49,#EB,#F1,#88,#07,#47,#EB,#CC,#88,#07,#47,#88,#27,#47,
#EB,#CB,#88,#07,#47,#88,#27,#47,#88,#1F,#83,#EF,#02,#EB,#C8})
poke4(clear_copy + 5, clear_copy + 202) -- @dest_vs
poke4(clear_copy + 25, clear_copy + 170) -- @source_vs
poke4(clear_copy + 52, clear_copy + 207) -- @dest_x
poke4(clear_copy + 58, clear_copy + 175) -- @source_x
poke4(clear_copy + 82, clear_copy + 154) -- @source_w
poke4(clear_copy + 109, clear_copy + 214) -- @dest_y
poke4(clear_copy + 115, clear_copy + 182) -- @source_y
poke4(clear_copy + 136, clear_copy + 165) -- @source_h
poke4(clear_copy + 149, loop_proc)
poke4(clear_copy + 197, clear_copy + 319) -- @source_add
poke4(clear_copy + 229, clear_copy + 325) -- @dest_add
poke4(clear_copy + 251, clear_copy + 288) -- @line1
poke4(clear_copy + 267, clear_copy + 288) -- @line1
poke4(clear_copy + 275, clear_copy + 311) -- @line2
poke4(clear_copy + 281, clear_copy + 430) -- @line3
poke4(clear_copy + 335, clear_copy + 175) -- @source_x
poke4(clear_copy + 341, clear_copy + 207) -- @dest_x
poke4(clear_copy + 347, clear_copy + 154) -- @source_w
poke4(clear_copy + 360, loop_proc)
mix_copy = allocate_proc({
#8B,#7B,#2F,#89,#3D,#83,#77,#4D,#00,#8B,#73,#33,#8B,#46,#1C,#85,#C0,#74,#03,#89,
#43,#33,#8B,#06,#A3,#63,#77,#4D,#00,#8B,#56,#0C,#8B,#43,#03,#8B,#4E,#04,#03,#46,
#14,#7D,#08,#01,#C2,#7E,#64,#29,#C1,#31,#C0,#A3,#88,#77,#4D,#00,#89,#0D,#68,#77,
#4D,#00,#F7,#D8,#03,#47,#10,#39,#C2,#7F,#02,#89,#D0,#83,#F8,#00,#7E,#45,#83,#C0,
#03,#A3,#53,#77,#4D,#00,#8B,#56,#10,#8B,#43,#07,#8B,#4E,#08,#03,#46,#18,#7D,#08,
#01,#C2,#7E,#2B,#29,#C1,#31,#C0,#A3,#8F,#77,#4D,#00,#89,#0D,#6F,#77,#4D,#00,#F7,
#D8,#03,#47,#14,#39,#C2,#7F,#02,#89,#D0,#83,#F8,#00,#7E,#0C,#A3,#5E,#77,#4D,#00,
#BD,#04,#00,#00,#00,#EB,#06,#FF,#25,#00,#00,#00,#00,#8B,#43,#37,#A3,#FC,#77,#4D,
#00,#A3,#7C,#78,#4D,#00,#A3,#8F,#78,#4D,#00,#A3,#9E,#78,#4D,#00,#A3,#B1,#78,#4D,
#00,#A3,#BD,#78,#4D,#00,#B9,#00,#00,#00,#00,#C1,#E9,#02,#74,#CF,#55,#BD,#00,#00,
#00,#00,#BE,#00,#00,#00,#00,#BB,#00,#00,#00,#00,#8B,#06,#BA,#00,#00,#00,#00,#C1,
#E8,#02,#FF,#56,#08,#29,#C8,#89,#DE,#A3,#11,#78,#4D,#00,#BF,#00,#00,#00,#00,#BB,
#00,#00,#00,#00,#8B,#07,#BA,#00,#00,#00,#00,#C1,#E8,#02,#FF,#57,#08,#8B,#1D,#88,
#77,#4D,#00,#8B,#15,#8F,#77,#4D,#00,#FF,#57,#0C,#29,#C8,#89,#DF,#A3,#17,#78,#4D,
#00,#F7,#DB,#31,#C0,#81,#E3,#03,#00,#00,#00,#29,#D9,#7D,#04,#01,#CB,#31,#C9,#89,
#1D,#EA,#77,#4D,#00,#88,#CB,#80,#E3,#03,#C1,#E9,#02,#75,#08,#01,#1D,#EA,#77,#4D,
#00,#89,#CB,#89,#0D,#09,#78,#4D,#00,#89,#1D,#6C,#78,#4D,#00,#EB,#00,#B9,#00,#00,
#00,#00,#85,#C9,#74,#15,#8A,#06,#84,#C0,#74,#0A,#8A,#27,#8A,#80,#00,#00,#00,#00,
#88,#07,#46,#47,#49,#75,#EB,#81,#C1,#00,#00,#00,#00,#75,#2F,#81,#C6,#00,#00,#00,
#00,#81,#C7,#00,#00,#00,#00,#4D,#75,#CB,#5D,#FF,#05,#68,#77,#4D,#00,#FF,#05,#88,
#77,#4D,#00,#FF,#0D,#53,#77,#4D,#00,#4D,#0F,#85,#1A,#FF,#FF,#FF,#FF,#25,#00,#00,
#00,#00,#8B,#16,#85,#D2,#74,#1D,#8B,#1F,#84,#D2,#75,#3D,#84,#F6,#75,#47,#C1,#CA,
#10,#84,#D2,#75,#51,#C1,#CB,#10,#84,#F6,#75,#5B,#C1,#CB,#10,#89,#1F,#83,#C6,#04,
#83,#C7,#04,#49,#75,#D4,#81,#C1,#00,#00,#00,#00,#74,#9D,#8A,#06,#84,#C0,#74,#0A,
#8A,#27,#8A,#80,#00,#00,#00,#00,#88,#07,#46,#47,#49,#EB,#E9,#88,#DC,#88,#D0,#84,
#F6,#8A,#98,#00,#00,#00,#00,#74,#B9,#88,#FC,#88,#F0,#C1,#CA,#10,#8A,#B8,#00,#00,
#00,#00,#84,#D2,#74,#AF,#C1,#CB,#10,#88,#D0,#88,#DC,#84,#F6,#8A,#98,#00,#00,#00,
#00,#74,#A5,#88,#F0,#88,#FC,#8A,#B8,#00,#00,#00,#00,#C1,#CB,#10,#89,#1F,#EB,#99})
poke4(mix_copy + 5, mix_copy + 235) -- @dest_vs
poke4(mix_copy + 25, mix_copy + 203) -- @source_vs
poke4(mix_copy + 52, mix_copy + 240) -- @dest_x
poke4(mix_copy + 58, mix_copy + 208) -- @source_x
poke4(mix_copy + 82, mix_copy + 187) -- @source_w
poke4(mix_copy + 109, mix_copy + 247) -- @dest_y
poke4(mix_copy + 115, mix_copy + 215) -- @source_y
poke4(mix_copy + 136, mix_copy + 198) -- @source_h
poke4(mix_copy + 149, loop_proc)
poke4(mix_copy + 157, mix_copy + 356) -- @table1
poke4(mix_copy + 162, mix_copy + 484) -- @table2
poke4(mix_copy + 167, mix_copy + 503) -- @table3
poke4(mix_copy + 172, mix_copy + 518) -- @table4
poke4(mix_copy + 177, mix_copy + 537) -- @table5
poke4(mix_copy + 182, mix_copy + 549) -- @table6
poke4(mix_copy + 230, mix_copy + 377) -- @source_add
poke4(mix_copy + 259, mix_copy + 240) -- @dest_x
poke4(mix_copy + 265, mix_copy + 247) -- @dest_y
poke4(mix_copy + 277, mix_copy + 383) -- @dest_add
poke4(mix_copy + 301, mix_copy + 338) -- @line1
poke4(mix_copy + 317, mix_copy + 338) -- @line1
poke4(mix_copy + 325, mix_copy + 369) -- @line2
poke4(mix_copy + 331, mix_copy + 468) -- @line3
poke4(mix_copy + 393, mix_copy + 208) -- @source_x
poke4(mix_copy + 399, mix_copy + 240) -- @dest_x
poke4(mix_copy + 405, mix_copy + 187) -- @source_w
poke4(mix_copy + 418, loop_proc)

-- isn't self modifying code awesome?
