-- palette.e
-- Palette routines
-- Pete Eberlein <xseal@harborside.com>
-- May 20, 1998

-- find_colors: converts {{r0,g0,b0},...,{rn,gn,bn}} to {c0,c1,...,cn}
--              using a given palette
-- choose_palette: selects which palette to use 

-- fade_palette: fades between two palettes, level = 0..64
--   fade_palette(0): fades to previous choose_palette
--   fade_palette(64): fades to current choose_palette
-- note: no effect if less than two palettes have been previously chosen
--       or if the two palettes are not the same size

-- set_palette: give a palette to set it,
--              or 0 uses the previous palette,
--              or 1 uses the current palette
                

include machine.e

atom squares
squares = allocate(1024)
for i = 0 to 255 do
    poke4(squares + i*4, power(i-and_bits(i*2,256),2))
end for

constant findcol = allocate(88)
poke(findcol,
   {#60,                    --    0: pusha
    #BA,#00,#00,#00,#00,    --    1: mov edx, source (2)
    #BF,#00,#00,#00,#00,    --    6: mov edi, dest (7)
    #B9,#00,#00,#00,#00,    --    B: mov ecx, count (12)
                            --   10: l1:
    #51,                    --   10: push ecx
    #BE,#00,#00,#00,#00,    --   11: mov esi, palette (18)
    #31,#C9,                --   16: xor ecx, ecx
    #BB,#A0,#86,#01,#00,    --   18: mov ebx, 100000
    #31,#C0,                --   1D: xor eax, eax
                            --   1F: l2:
    #AC,                    --   1F: lodsb
    #2A,#02,                --   20: sub al, [edx]
    #8B,#2C,#85,#00,#00,#00,#00,--   22: mov ebp, dword ptr [eax*4 + dword squares] (37)
    #AC,                    --   29: lodsb
    #2A,#42,#01,            --   2A: sub al, [edx+1]
    #03,#2C,#85,#00,#00,#00,#00,--   2D: add ebp, dword ptr [eax*4 + dword squares] (48)
    #AC,                    --   34: lodsb
    #2A,#42,#02,            --   35: sub al, [edx+2]
    #03,#2C,#85,#00,#00,#00,#00,--   38: add ebp, dword ptr [eax*4 + dword squares] (59)
    #39,#DD,                --   3F: cmp ebp, ebx
    #7D,#04,                --   41: jge l3
    #89,#EB,                --   43: mov ebx, ebp
    #88,#0F,                --   45: mov [edi], cl
                            --   47: l3:
    #FE,#C1,                --   47: inc cl
    #80,#F9,#00,            --   49: cmp cl, byte palsize (75)
    #75,#D1,                --   4C: jne l2
    #59,                    --   4E: pop ecx
    #47,                    --   4F: inc edi
    #42,                    --   50: inc edx
    #42,                    --   51: inc edx
    #42,                    --   52: inc edx
    #49,                    --   53: dec ecx
    #75,#BA,                --   54: jnz l1
    #61,                    --   56: popa
    #C3}                    --   57: ret
    )
   
atom mem_palette, mem_palsize, old_palette
mem_palette = 0
old_palette = 0

global function find_colors(sequence colors3)
    sequence result
    atom source, dest
    source = allocate(3 * length(colors3))
    if source = 0 then return {} end if
    dest = allocate(length(colors3))
    if dest = 0 then return {} end if
    
    for i = 1 to length(colors3) do
        poke(source+i*3-3, colors3[i])
    end for

    poke4(findcol + 2, source)
    poke4(findcol + 7, dest)
    poke4(findcol + 12, length(colors3))
    poke4(findcol + 18, mem_palette)
    poke4(findcol + 37, squares)
    poke4(findcol + 48, squares)
    poke4(findcol + 59, squares)
    poke(findcol + 75, mem_palsize)

    call(findcol)
    free(source)
    result = peek({dest, length(colors3)})
    free(dest)
    return result
end function

global procedure choose_palette(sequence pal)
    if old_palette then
        free(old_palette)
        old_palette = 0
    end if
    if mem_palette then 
        if mem_palsize = length(pal) then
            old_palette = mem_palette
        else
            free(mem_palette)
        end if
    end if
    mem_palette = allocate(3 * length(pal))
    mem_palsize = length(pal)
    for i = 1 to length(pal) do
        poke(mem_palette+i*3-3, pal[i])
    end for
end procedure


constant setpal = allocate(25)
poke(setpal,
   {#60,                    --    0: pusha
    #66,#BA,#C8,#03,        --    1: mov dx, 968
    #31,#C0,                --    5: xor eax, eax
    #EE,                    --    7: out dx, al
    #66,#FF,#C2,            --    8: inc dx
    #BE,#00,#00,#00,#00,    --    B: mov esi, mem_palette (12)
    #B9,#00,#00,#00,#00,    --   10: mov ecx, mem_palsize (17)
    #F3,#6E,                --   15: rep outsb
    #61,                    --   17: popa
    #C3}                    --   18: ret
   )


global procedure set_palette(object pal)
    atom mem_pal
    if sequence(pal) then
        mem_pal = allocate(3 * length(pal))
        poke4(setpal + 12, mem_pal)
        poke4(setpal + 17, 3 * length(pal))
        for i = 1 to length(pal) do
            poke(mem_pal+i*3-3, pal[i])
        end for
        call(setpal)
        free(mem_pal)
    else
        poke4(setpal + 17, mem_palsize * 3)
        if pal = 0 and old_palette then
            poke4(setpal + 12, old_palette)
            call(setpal)
        elsif pal = 1 and mem_palette then
            poke4(setpal + 12, mem_palette)
            call(setpal)
        end if
    end if
end procedure


constant fader = allocate(49)
poke(fader, 
   {#60,                    --    0: pusha
    #BA,#C8,#03,#00,#00,    --    1: mov edx, 968
    #31,#C0,                --    6: xor eax, eax
    #EE,                    --    8: out dx, al
    #42,                    --    9: inc edx
    #31,#C9,                --    A: xor ecx, ecx
    #B7,#00,                --    C: mov bh, fade (13)
    #31,#C0,                --    E: l1:  xor eax, eax
    #8A,#99,#00,#00,#00,#00,--   10: mov bl, byte ptr [ecx + dword pal1] (18)
    #8A,#81,#00,#00,#00,#00,--   16: mov al, byte ptr [ecx + dword pal2] (24)
    #28,#D8,                --   1C: sub al, bl
    #F6,#EF,                --   1E: imul bh
    #C1,#E8,#06,            --   20: shr eax, 6
    #00,#D8,                --   23: add al, bl
    #EE,                    --   25: out dx, al
    #41,                    --   26: inc ecx
    #81,#F9,#00,#00,#00,#00,--   27: cmp ecx, dword palsize (41)
    #75,#DF,                --   2D: jne l1
    #61,                    --   2F: popa
    #C3}                    --   30: ret
   )

global procedure fade_palette(integer level)  -- level: 0 .. 64
    if old_palette and mem_palette then
        poke(fader + 13, level)
        poke4(fader + 18, old_palette)
        poke4(fader + 24, mem_palette)
        poke4(fader + 41, mem_palsize * 3)
        call(fader)
    end if
end procedure

global function gray_scale(sequence s)
    for i = 1 to length(s) do
        s[i] = repeat(floor(s[i][1]*0.39 + s[i][2]*0.41 + s[i][3]*0.2), 3)
    end for    
    return s
end function

global procedure border_color(integer color)
    atom bc
    bc = allocate(16)
    if bc then
        poke(bc, 
           {#60,                    --    0: pusha
            #66,#BA,#DA,#03,        --    1: mov dx, #3DA
            #EC,                    --    5: in al, dx
            #B2,#C0,                --    6: mov dl, #C0
            #B0,#31,                --    8: mov al, #31
            #EE,                    --    A: out dx, al
            #B0,color,              --    B: mov al, yourcolor (12)
            #EE,                    --    D: out dx, al
            #61,                    --    E: popa
            #C3})                   --    F: ret
        call(bc)
        free(bc)
    end if
end procedure

global constant default_palette =
{{0,0,0},{0,0,42},{0,42,0},{0,42,42},{42,0,0},{42,0,42},{42,21,0},{42,
42,42},{21,21,21},{21,21,63},{21,63,21},{21,63,63},{63,21,21},{63,21,
63},{63,63,21},{63,63,63},{0,0,0},{5,5,5},{8,8,8},{11,11,11},{14,14,
14},{17,17,17},{20,20,20},{24,24,24},{28,28,28},{32,32,32},{36,36,36},
{40,40,40},{45,45,45},{50,50,50},{56,56,56},{63,63,63},{0,0,63},{16,
0,63},{31,0,63},{47,0,63},{63,0,63},{63,0,47},{63,0,31},{63,0,16},{63,
0,0},{63,16,0},{63,31,0},{63,47,0},{63,63,0},{47,63,0},{31,63,0},{16,
63,0},{0,63,0},{0,63,16},{0,63,31},{0,63,47},{0,63,63},{0,47,63},{0,
31,63},{0,16,63},{31,31,63},{39,31,63},{47,31,63},{55,31,63},{63,31,
63},{63,31,55},{63,31,47},{63,31,39},{63,31,31},{63,39,31},{63,47,31},
{63,55,31},{63,63,31},{55,63,31},{47,63,31},{39,63,31},{31,63,31},{31,
63,39},{31,63,47},{31,63,55},{31,63,63},{31,55,63},{31,47,63},{31,39,
63},{45,45,63},{49,45,63},{54,45,63},{58,45,63},{63,45,63},{63,45,58},
{63,45,54},{63,45,49},{63,45,45},{63,49,45},{63,54,45},{63,58,45},{63,
63,45},{58,63,45},{54,63,45},{49,63,45},{45,63,45},{45,63,49},{45,63,
54},{45,63,58},{45,63,63},{45,58,63},{45,54,63},{45,49,63},{0,0,28},
{7,0,28},{14,0,28},{21,0,28},{28,0,28},{28,0,21},{28,0,14},{28,0,7},
{28,0,0},{28,7,0},{28,14,0},{28,21,0},{28,28,0},{21,28,0},{14,28,0},
{7,28,0},{0,28,0},{0,28,7},{0,28,14},{0,28,21},{0,28,28},{0,21,28},
{0,14,28},{0,7,28},{14,14,28},{17,14,28},{21,14,28},{24,14,28},{28,
14,28},{28,14,24},{28,14,21},{28,14,17},{28,14,14},{28,17,14},{28,21,
14},{28,24,14},{28,28,14},{24,28,14},{21,28,14},{17,28,14},{14,28,14},
{14,28,17},{14,28,21},{14,28,24},{14,28,28},{14,24,28},{14,21,28},{14,
17,28},{20,20,28},{22,20,28},{24,20,28},{26,20,28},{28,20,28},{28,20,
26},{28,20,24},{28,20,22},{28,20,20},{28,22,20},{28,24,20},{28,26,20},
{28,28,20},{26,28,20},{24,28,20},{22,28,20},{20,28,20},{20,28,22},{20,
28,24},{20,28,26},{20,28,28},{20,26,28},{20,24,28},{20,22,28},{0,0,
16},{4,0,16},{8,0,16},{12,0,16},{16,0,16},{16,0,12},{16,0,8},{16,0,
4},{16,0,0},{16,4,0},{16,8,0},{16,12,0},{16,16,0},{12,16,0},{8,16,0},
{4,16,0},{0,16,0},{0,16,4},{0,16,8},{0,16,12},{0,16,16},{0,12,16},{0,
8,16},{0,4,16},{8,8,16},{10,8,16},{12,8,16},{14,8,16},{16,8,16},{16,
8,14},{16,8,12},{16,8,10},{16,8,8},{16,10,8},{16,12,8},{16,14,8},{16,
16,8},{14,16,8},{12,16,8},{10,16,8},{8,16,8},{8,16,10},{8,16,12},{8,
16,14},{8,16,16},{8,14,16},{8,12,16},{8,10,16},{11,11,16},{12,11,16},
{13,11,16},{15,11,16},{16,11,16},{16,11,15},{16,11,13},{16,11,12},{16,
11,11},{16,12,11},{16,13,11},{16,15,11},{16,16,11},{15,16,11},{13,16,
11},{12,16,11},{11,16,11},{11,16,12},{11,16,13},{11,16,15},{11,16,16},
{11,15,16},{11,13,16},{11,12,16},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,
0,0},{0,0,0},{0,0,0},{0,0,0}}

constant rotateproc = allocate(72)
poke(rotateproc, {
    #60,                    --    0: pusha
    #BA,#C8,#03,#00,#00,    --    1: mov edx, 968
    #B0,#00,                --    6: mov al, first_color (7)
    #EE,                    --    8: out dx, al
    #BE,#00,#00,#00,#00,    --    9: mov esi, palette (10)
    #89,#F7,                --    E: mov edi, esi
    #66,#FF,#C2,            --   10: inc dx
    #FC,                    --   13: cld
    #8A,#3E,                --   14: mov bh, [esi]
    #46,                    --   16: inc esi
    #8A,#1E,                --   17: mov bl, [esi]
    #46,                    --   19: inc esi
    #C1,#E3,#08,            --   1A: shl ebx, 8
    #8A,#1E,                --   1D: mov bl, [esi]
    #46,                    --   1F: inc esi
    #B9,#00,#00,#00,#00,    --   20: mov ecx, count (33)
                            --   25: loop1:
    #8A,#06,                --   25:     mov al, [esi]
    #A4,                    --   27: movsb
    #EE,                    --   28: out dx, al
    #8A,#06,                --   29: mov al, [esi]
    #A4,                    --   2B: movsb
    #EE,                    --   2C: out dx, al
    #8A,#06,                --   2D: mov al, [esi]
    #A4,                    --   2F: movsb
    #EE,                    --   30: out dx, al
    #49,                    --   31: dec ecx
    #75,#F1,                --   32: jnz loop1
    #89,#D8,                --   34: mov eax, ebx
    #C1,#C8,#10,            --   36: ror eax, 16
    #EE,                    --   39: out dx, al
    #AA,                    --   3A: stosb
    #C1,#C0,#08,            --   3B: rol eax, 8
    #EE,                    --   3E: out dx, al
    #AA,                    --   3F: stosb
    #C1,#C0,#08,            --   40: rol eax, 8
    #EE,                    --   43: out dx, al
    #AA,                    --   44: stosb
    #61,                    --   45: popa
    #C3})                   --   46: ret

global procedure rotate_palette(integer first, integer last)
    if mem_palette then
        poke(rotateproc + 7, first)
        poke4(rotateproc + 10, mem_palette + first*3)
        poke4(rotateproc + 33, last-first)
        call(rotateproc)
    end if
end procedure
